(*************************************************************************)
(*                                                                       *)
(*                                OCaPIC                                 *)
(*                                                                       *)
(*            Benoit Vaugon, Université Pierre et Marie Curie            *)
(*                                                                       *)
(*    Ce fichier est distribué sous les termes de la licence CeCILL-B    *)
(*    décrite dans le fichier ../LICENCE.                                *)
(*                                                                       *)
(*************************************************************************)

(** Character type LCD display communication interface. *)

(** Standard PIC signature. *)
module type P =
sig
  type bit
    (** Type of bits of Special Function Registers. *)

  type reg
    (** Type of Special Function Registers. *)

  val write_reg : reg -> int -> unit
    (** [write_reg reg value] writes value in the Special Function Register reg.
	value should be between 0 and 255. *)

  val read_reg : reg -> int
    (** Reads value of a Special Function Register value.
	Return value between 0 and 255. *)

  val set_bit : bit -> unit
    (** Sets a Special Function Register bit. *)

  val clear_bit : bit -> unit
    (** Clears a Special Function Register bit. *)

  val tris_of_port : reg -> reg
    (** Convertion from a port to its configuration register. *)

  val tris_of_pin : bit -> bit
    (** Convertion from a pin to its configuration bit. *)

end

(** Inforamtions about connection between the PIC and the LCD display. *)
module type S =
sig
  module Pic : P
    (** Pic used. *)

  val is_8_bits : bool
    (** Bus size: true -> 8 ; false -> 4. *)

  val e : Pic.bit
    (** Pin of the pic connected to the [e] entry of the LCD display. *)

  val rs : Pic.bit
    (** Pin of the pic connected to the [rs] entry of the LCD display. *)

  val rw : Pic.bit
    (** Pin of the pic connected to the [rw] entry of the LCD display. *)

  val port : Pic.reg
    (** Port of the pic connected to the bus of the LCD display.
        In case of 4 bit bus, the 4th highest bits of the port are used. *)

end

(** Connects the PIC to the LCD display.
    Takes informations about connection in the module C. *)
module Connect :
  functor (C : S) ->
    sig
      val send : int -> unit
        (** Sends data or instruction to the LCD display. *)

      val receive : unit -> int
        (** Receives data or address from the LCD display. *)

      val send_data : int -> unit
        (** Sends data to the LCD display. *)

      val receive_data : unit -> int
        (** Receives data from the LCD display. *)

      val send_instr : int -> unit
        (** Sends instruction to the LCD display. *)

      val is_busy : unit -> bool
        (** Reads the busy flag of the LCD display. *)

      val get_addr : unit -> int
        (** Reads the current cursor address on the LCD display. *)

      val set_cgram : int -> unit
        (** Sets the CGRAM address of the LCD display. *)

      val set_ddram : int -> unit
        (** Sets the DDRAM address of the LCD display (cursor moving). *)

      val clear : unit -> unit
        (** Clears the LCD display. *)

      val home : unit -> unit
        (** Returns cursor and display to their initial positions. *)

      val write_string : string -> unit
        (** Writes string on the LCD display at the current cursor position.
            Moves cursor. *)

      type direction
        (** Direction of cursor or display shift. *)

      val left : direction
        (** Left direction of cursor or display shift. *)

      val right : direction
        (** Right direction of cursor or display shift. *)

      val shift_cursor : direction -> unit
        (** Shifts cursor of 1 character. *)

      val shift_display : direction -> unit
        (** Shifts display of 1 character. *)

      type mode = Cursor_left | Cursor_right | Display_left | Display_right
        (** Writing mode: cursor or display moving and direction setting. *)

      type state = On | Off
        (** State of cursor or LCD display. *)

      type lmode = One | Two
        (** Line mode. *)

      type font = F5x8 | F5x10
        (** Characters font. *)

      val config : mode:mode -> disp:state -> cursor:state -> blink:state ->
        lmode:lmode -> font:font -> unit
        (** Configures the LCD display.
            Should be called after the LCD initialisation. *)

      val init : unit -> unit
        (** Initialises the LCD display.
            Should be called once on the LCD power on. *)

    end
