(*************************************************************************)
(*                                                                       *)
(*                                OCaPIC                                 *)
(*                                                                       *)
(*            Benoit Vaugon, Université Pierre et Marie Curie            *)
(*                                                                       *)
(*    Ce fichier est distribué sous les termes de la licence CeCILL-B    *)
(*    décrite dans le fichier ../../LICENCE.                             *)
(*                                                                       *)
(*************************************************************************)

exception Exn of string

type elem =
  | String of string
  | Float of float
  | Floats of float array
  | Int of int
  | Block of elem array * int
  | Closure of elem array
  | Out_of_heap of int
;;

type mem_item =
  | Val of int
  | Addr of int
  | Flp of float
  | Header of int * int
  | Charchar of int * int
;;

let parse ic index =
  let (offset, _) =
    try
      Index.find_section index Index.Data
    with Not_found ->
      raise (Exn "code section not found")
  in
    seek_in ic offset;
    let (tbl : Obj.t array) = input_value ic in
    let rec elem_of_obj obj =
      let tag = Obj.tag obj in
	if tag = Obj.string_tag then String (Obj.obj obj)
	else if tag = Obj.double_tag then Float (Obj.obj obj)
	else if tag = Obj.double_array_tag then Floats (Obj.obj obj)
	else if tag = Obj.int_tag then Int (Obj.obj obj)
	else if tag = Obj.out_of_heap_tag then Out_of_heap (Obj.obj obj)
	else
	  let size = Obj.size obj in
	  let tab =
	    Array.init size (fun i -> elem_of_obj (Obj.field obj i))
	  in
	    if tag = Obj.closure_tag then Closure tab
	    else Block (tab, tag)
    in
      Array.map elem_of_obj tbl
;;

let print oc data =
  let rec print_elem elem =
    match elem with
      | String str -> Printf.fprintf oc "\"%s\"" str
      | Float f    -> Printf.fprintf oc "%f" f
      | Floats tbl ->
	  Printf.fprintf oc "[|";
	  Array.iter (Printf.fprintf oc " %f ") tbl;
	  Printf.fprintf oc "|]";
      | Int n      -> Printf.fprintf oc "%d" n
      | Block (tab, tag) ->
	  Printf.fprintf oc "[%d|" tag;
	  Array.iter
	    (fun e -> output_char oc ' ' ; print_elem e ; output_char oc ' ')
	    tab;
	  Printf.fprintf oc "|%d]" tag;
      | Closure tab ->
	  Printf.fprintf oc "Closure [|";
	  Array.iter
	    (fun e -> output_char oc ' ' ; print_elem e ; output_char oc ' ')
	    tab;
	  Printf.fprintf oc "|]";
      | Out_of_heap p -> Printf.fprintf oc "@0x%08x" p;
  in
  let print_elem i elem =
    Printf.fprintf oc "%-3d   " i;
    print_elem elem;
    output_char oc '\n';
  in
    Array.iteri print_elem data;
;;

let to_init data =
  let stack = Queue.create () in
  let heap = Queue.create () in
  let push_stack x = Queue.push x stack in
  let push_heap x = Queue.push x heap in
  let check_float f =
    let (_, n) = frexp f in
      if n < -32 || n > 31 then
	raise (Exn (Printf.sprintf "float out of range: %f" f));
  in
  let rec export e =
    match e with
      | Int i ->
	  if i < -16384 || i > 16383 then
	    raise (Exn (Printf.sprintf
			  "Integer out of bounds in global data: %d" i));
	  Val i
      | String str ->
	  let str_len = String.length str in
	  let blk_len = str_len / 2 + 1 in
	    if blk_len > 255 then
	      raise (Exn (Printf.sprintf
			    "Too large string in global data: \"%s\"" str));
	    push_heap (Header (blk_len, Obj.string_tag));
	    let res = Addr (2*Queue.length heap + Constants.heap1_anchor) in
	      for i = 0 to blk_len - 2 do
		let ci = int_of_char str.[2 * i] in
		let csi = int_of_char str.[2 * i + 1] in
		  push_heap (Charchar (ci, csi))
	      done;
	      if str_len mod 2 = 0 then
		push_heap (Charchar (0, 1))
	      else
		push_heap (Charchar (int_of_char str.[str_len - 1], 0));
	      res
      | Block (tab, tag) ->
	  let size = Array.length tab in
	    if size = 0 then
	      Addr Constants.atom0_location
	    else (
	      if size > 255 then
		raise (Exn (Printf.sprintf
			      "Too big block in global data: %d > 255" size));
	      let body = Array.make size (Val 42) in
		for i = 0 to size - 1 do body.(i) <- export tab.(i) done;
		push_heap (Header (size, tag));
		let res = Addr (2*Queue.length heap + Constants.heap1_anchor) in
		  for i = 0 to size - 1 do push_heap body.(i) done;
		  res
	    )
      | Float f ->
	  check_float f;
	  push_heap (Header (1, Obj.double_tag));
	  let res = Addr (2*Queue.length heap + Constants.heap1_anchor) in
	    push_heap (Flp f);
	    res
      | Floats tab ->
	  let size = Array.length tab in
	    if size = 0 then
	      raise (Exn "float array of size 0")
	    else (
	      push_heap (Header (size, Obj.double_array_tag));
	      let res = Addr (2*Queue.length heap + Constants.heap1_anchor) in
		for i = 0 to size - 1 do
		  check_float tab.(i);
		  push_heap (Flp tab.(i));
		done;
		res
	    )
      | Closure _ -> raise (Exn "closure in global data unsupported")
      | Out_of_heap _ -> raise (Exn "\"out of heap\" in global data")
  in
    Array.iter (fun e -> push_stack (export e)) data;
    (stack, heap)
;;

let export oc data =
  let (stack, heap) = to_init data in
  let cpt = ref 0 in
  let put_byte byte =
    if !cpt land 0b111 = 0 then
      output_string oc "\n        db      "
    else
      output_string oc ", ";
    Printf.fprintf oc "0x%02x" byte;
    incr cpt;
  in
  let export e =
    match e with
      | Val n ->
	  put_byte (((n land 0x7F) lsl 1) lor 1);
	  put_byte (n lsr 7);
      | Addr n ->
	  put_byte (n land 0xFF);
	  put_byte (n lsr 8);
      | Flp f ->
	  let (x, e) = frexp f in
	  let rec d i a r =
	    if i < 0 then
	      r
	    else if a < 0.5 then
	      d (pred i) (a *. 2.) r
	    else
	      d (pred i) ((a -. 0.5) *. 2.) (r lor (1 lsl i))
	  in
	  let s = if f < 0. then 1 else 0 in
	  let m = d 10 (abs_float x) 0 land 0x3FF in
	  let b1 = m land 0xFF in
	  let b2 = (s lsl 7) lor ((m lsr 8) lsl 5) lor (e land 0b11111) in
	    put_byte b1;
	    put_byte b2;
      | Header (size, tag) ->
	  put_byte tag;
	  put_byte size;
      | Charchar (c1, c2) ->
	  put_byte c1;
	  put_byte c2;
  in
    Printf.fprintf oc "caml_globals_init_stack:";
    Queue.iter export stack;
    cpt := 0;
    Printf.fprintf oc "\ncaml_globals_init_heap:";
    Queue.iter export heap;
    Printf.fprintf oc "\ncaml_globals_init_end:\n\n";
;;
